#ifndef INDIVIDUDISCRIMINANT_H_
#define INDIVIDUDISCRIMINANT_H_

/* include pour developpement */
//#include "../../../../CPT/include/dae.h"
//include pour la livraison
#include "../../../CPT/include/dae.h"
#include "../cptUtilities/cpt.h"
#include <sstream>
#include <stdexcept>
std::string resultFile;
using namespace std;
/*! @class VariablesGlobales
 * @file eoCPTEO_V1.h
 * @author Jacques Bibai
 * @brief Ensembles de parametres utilises dans l'application
 * Cette classe permet de stocker toutes les variables utilisees dans l'application.  
 * Ces variables sont des parametres de CPT et aussi ceux permettant le parametrage de EO. 
 */
unsigned int initGlobal(eoParser & _parser){
	return (unsigned int)  _parser.getORcreateParam(1800, "runTime", "run duration", 't', "DAE parameters").value();
}

class VariablesGlobales : public eoPrintable 
{
	/**
	 *  Attributs globales ou ensemble des variables globales utilisees
	 */
	private :
		  unsigned debug; // variable permettant d'afficher des commentaires sur la sortie standard 
		  unsigned int minSize;	// variable representant la taille minimun du nombre de stations
		  unsigned int maxSize;	// variable representant la taille maximun du nombre de stations
		  unsigned int maxDist;	 // variable representant le nombre de fluent par stations
		  double pAddRemovePredicate; // variable representant la probabilite des changements suplementaire	   
		  double avgNbAdditionalAtChange ; // 
		  double atMutRatePerGoal; // 
		  unsigned nbPred; // variable representant le nombre de predicats pris en compte lors de la construction d'un individu
		  std::string maxb; // variable representant le nombre de backtracks autorises dans CPT
		  std::string NameDomaine; // variable representant le nom du domaine definissant le probleme
		  std::string NameObjects;// variable representant le nom des objets du probleme
		  std::vector<string> NamePredicates; // liste des predicats
		  double penalite; // variable definissant la penalite pour la fitness
		  unsigned distance;
		  std::vector<unsigned> InitialState;
		  double pMeanAveragePerMutation;
		  int maxbks;
		  unsigned int 	runTime;
		   double precision;
	public :
		std::vector<unsigned> problemGoal;
		std::vector<unsigned> problemInit;
		std::vector<unsigned> mutexCorrespondance;
		unsigned stationMax;
		/**
		 * constructeur par defaut
		 */
		VariablesGlobales(void)
		{
			
		}
		
		/**
		 * @brief affichage des attributes spaces de CPT
		 * @fn void printAttributeSpaces()
		 * Methode permettant d'afficher les attributes spaces
		 */
		 void printAttributeSpaces()
		 {
			  int i, j, k;
			
			  cout << "\nATTRIBUTE SPACES\n\n";
			  CPTPredicateArray preds = cptGetPredicates();
			  for (i = 0; i < preds.nb; i++) {
			    CPTPredicate pred = preds.elems[i];
			    cout << "PREDICATE %s :\n" << cptGetPredicateName(pred);
			    // pour chaque argument possible d'un predicat : affichage
			    // de la liste des objets possibles
			    for (j = 0; j < cptGetArity(pred); j++) {
			      cout << "   ARG " << j << " :";
			      CPTTermArray terms = cptGetPossibleTerms(pred, j);
			      for (k = 0; k < terms.nb; k++)
				cout << " " << cptGetTermName(terms.elems[k]);
			      cout << "\n";
			    }
			  }
			}
			
			void help(eoParser & _parser){
				debug = _parser.getORcreateParam(0, "debug", "Debug mode ", 'd', "DAE parameters").value();
		   
				//   _parser.getORcreateParam(unsigned(1), "minSize", "Minimal size at init", 'm', "DAE parameters");
				     _parser.getORcreateParam(unsigned(3), "maxSize", "Maximal size at init", 'l', "DAE parameters").value();
				   _parser.getORcreateParam(unsigned(3), "maxDist", "Maximal number of predicate per station", 'D', "DAE parameters").value();
				   _parser.getORcreateParam(unsigned(100), "distance", "Number atoms per station", 'a', "DAE parameters").value(); 
				    _parser.getORcreateParam(0.2, "pAddRemovePredicate", "Probability to change the 'present' state in init and mutation", '\0', "DAE parameters").value();
				     _parser.getORcreateParam(0.2, "pMeanAveragePerMutation", "Mean probability to change fluent of the  state in mutation", 'A', "DAE parameters").value();
					_parser.getORcreateParam(1e+60, "penalite", "Penality", 'p', "DAE parameters").value();
				     _parser.getORcreateParam(20, "stationMax", "nombre de stations maximum autorisees", 'm', "DAE parameters").value();
				  	 _parser.getORcreateParam(1800, "runTime", "run duration", 't', "DAE parameters").value();
				   	 _parser.getORcreateParam(std::string("result"), "fileResult", "name of output file", 'r', "Output").value();
					//_parser.getORcreateParam(unsigned(0), "nbPred", "nombre de predicat dans variables globales", '\0', "DAE parameters").value();
				   //_parser.getORcreateParam(std::string("2000"), "maxb", "nombre maximum de backtracks autorises", 'b', "DAE parameters").value();
				   _parser.getORcreateParam(std::string("domain.pddl"), "o", "name of domain file ", 'o', "Input").value();
				   _parser.getORcreateParam(std::string("object.pddl"), "f", "name of objects file", 'f', "Input").value();
				    _parser.getORcreateParam(0.1, "atMutRatePerGoal", "Mutation Rate per Patial Goal", '\0', "DAE parameters").value();
				    _parser.getORcreateParam(std::string("at/"), "liste", "liste des predicats", '\0', "DAE parameters").value(); 
				   _parser.createParam(35.0, "mut1Rate", "Relative rate for atoms modification mutation", '1', "Variation Operators").value();
				   _parser.createParam(35.0, "mut2Rate", "Relative rate for add station mutation", '2', "Variation Operators").value();
				   _parser.createParam(3.0, "mut3Rate", "Relative rate for del station mutation", '3', "Variation Operators").value();
				  _parser.createParam(7.0, "mut4Rate", "Relative rate for del atoms mutation", '4', "Variation Operators").value();
				   _parser.createParam(0.3, "pCross", "Probability of Crossover", 'c', "Variation Operators" ).value();
				    _parser.createParam(0.8, "pMut", "Probability of Mutation", 'M', "Variation Operators" ).value();
				   _parser.getORcreateParam(0.00001, "precision", "Validator precision", 'v', "DAE parameters").value();
			}
			void printDiscriminantAttribute(){
				int i,j,k;
				CPTPredicateArray preds = cptGetPredicates();
				std::vector<int> elemts_nb;
				std::vector<int> elemts;
				std::vector<int> elemts2;
				std::vector<int> sort_vec;
				std::vector<string> pred_name;
				std::vector<string> pred_name2;
				std::vector<int> elemts_nb2;
				std::vector<int> sort_vec2;
				for(i=0;i<preds.nb; i++) {
					CPTPredicate pred = preds.elems[i];
					if(cptGetArity(pred)==2){
						pred_name.push_back(cptGetPredicateName(pred));
						elemts.push_back(i);
						k=1;
						for (j = 0; j < cptGetArity(pred); j++) {
							CPTTermArray terms = cptGetPossibleTerms(pred, j);
							k=k*terms.nb;
						}
						elemts_nb.push_back(k);
						sort_vec.push_back(k);
					}
					if(cptGetArity(pred)==1){
						pred_name2.push_back(cptGetPredicateName(pred));
						elemts2.push_back(i);
						k=1;
						for (j = 0; j < cptGetArity(pred); j++) {
							CPTTermArray terms = cptGetPossibleTerms(pred, j);
							k=k*terms.nb;
						}
						elemts_nb2.push_back(k);
						sort_vec2.push_back(k);
					}
				}
				if(pred_name2.size()!=0)sort(sort_vec2.begin(), sort_vec2.end());
				if(pred_name.size()!=0)sort(sort_vec.begin(), sort_vec.end());
				int min;
				j=0;
				nbPred=0;
				distance=1;
				maxSize=0;
				if(pred_name.size()!=0){
					min=sort_vec[0];
					for(i=0;i<(int)elemts_nb.size();i++){
						if(elemts_nb[i]==min){
							j=i;
							break;
						}
					}
					cout << "\n Discriminant predicate 2------------------------ " << pred_name[j] << " \n";
					nbPred=1;
					NamePredicates.push_back(pred_name[j]);
					CPTPredicate pred = preds.elems[elemts[j]];
					CPTTermArray terms = cptGetPossibleTerms(pred, 0);
					maxSize=terms.nb;
					distance=terms.nb<3?terms.nb:3;
					if(pred_name.size()>2){
						min=sort_vec[sort_vec.size()/2];
						for(i=0;i<(int)elemts_nb.size();i++){
							if(elemts_nb[i]==min && j!=i){
								j=i;
								break;
							}
						}
						nbPred+=1;
						NamePredicates.push_back(pred_name[j]);
				 		pred = preds.elems[elemts[j]];
				 		terms = cptGetPossibleTerms(pred, 0);
				 		maxSize+=terms.nb;
						cout << "\n Discriminant predicate 2 ------------------------ " << pred_name[j] << " \n";
					}
				}
				
				if(pred_name2.size()!= 0){
					nbPred+=1;
					
					if(pred_name2.size()>0)min=sort_vec2[sort_vec2.size()-1];
					j=0;
					for(i=0;i<(int)elemts_nb2.size();i++){
						if(elemts_nb2[i]==min){j=i;break;}
					}
					cout << "\n Discriminant predicate 1------------------------ " << pred_name2[j] << " \n";
					NamePredicates.push_back(pred_name2[j]);
					CPTPredicate pred = preds.elems[elemts2[j]];
					CPTTermArray terms = cptGetPossibleTerms(pred, 0);
					maxSize+=terms.nb;
					if(pred_name.size()==0)distance=terms.nb<3?terms.nb:3;
					if(pred_name2.size()>1 && pred_name.size()==0)
					{	min=sort_vec2[sort_vec2.size()/2];
						for(i=0;i<(int)elemts_nb2.size();i++){
							if(elemts_nb2[i]==min && j!=i ){
								j=i;
								break;
							}
						}
						NamePredicates.push_back(pred_name2[j]);
					 	pred = preds.elems[elemts2[j]];
						terms = cptGetPossibleTerms(pred, 0);
						maxSize+=terms.nb;
						nbPred+=1;
						cout << "\n Discriminant predicate 1------------------------ " << pred_name2[j] << " \n";
					}
				}
				if(NamePredicates.size()==0){
					distance=3;
					for(i=0;i<preds.nb; i++) {
						CPTPredicate pred = preds.elems[i];
						if(cptGetArity(pred)==0){
							NamePredicates.push_back(cptGetPredicateName(pred));
						}
					}	
				}
				if(maxSize>stationMax)maxSize=stationMax;
				
			}
			
			/**
			 * @brief affichage d'une liste de fluents
			 * @fn void printFluentArray(CPTFluentArray)
			 * Methode permettant d'afficher une liste de fluents
			 * @param CPTFluentArray fluents : table de fluents
			 */
			void printFluentArray(CPTFluentArray fluents)
				{
				  int i, j;
				  for (i = 0; i < fluents.nb; i++) {
				    CPTFluent fluent = fluents.elems[i];
				    CPTPredicate pred = cptGetPredicate(fluent);
				    // affichage d'un fluent
				    cout << "FLUENT " << cptGetFluentName(fluent) 
					 << " :\n   NUM " << cptGetIndexFromFluent(fluent) 
					 << "\n   PREDICATE " << cptGetPredicateName(pred) 
					 << "\n   OBJECTS";
				    // affichage des termes
				    CPTTermArray terms = cptGetTerms(fluent);
				    for (j = 0; j < terms.nb; j++)
				      cout << " " << cptGetTermName(terms.elems[j]);
				    cout << "\n";
				    // recuperer le fluent a partir d'un predicat et d'un tableau de termes
				    CPTFluent fluent2 = cptGetFluent(pred, terms);
				    cout << "   GETFLUENT : " << cptGetFluentName(fluent2) << "\n";
				    // recuperer un fluent a  partir d'un numero
				    CPTFluent fluent3 = cptGetFluentFromIndex(cptGetIndexFromFluent(fluent));
				    cout << "   GETFLUENTFROMNUM : " << cptGetFluentName(fluent3) << "\n";
				  }
				}
				
			/**
			 * @brief initialise toutes les variables et CPT 
			 *  @fn void initialisation(eoParser &)
			 * Initialisation de tous les parametres (attributs) d'une instance de la classe variable
			 * globale a partir d'un fichier
			 * @param eoParser & _parser : instance de la classe eoParser
			 */
		
			void initialisation (eoParser & _parser)
			{
					// initialisation de debug
					debug = _parser.getORcreateParam(0, "debug", "Debug mode ", 'd', "DAE parameters").value();
		   
				   // eoValueParam<unsigned> & minSizeParam = _parser.getORcreateParam(unsigned(1), "minSize", "Minimal size at init", 'm', "DAE parameters");
				    //initialisation de maxSize    
				    maxSize = _parser.getORcreateParam(unsigned(3), "maxSize", "Maximal size at init", 'l', "DAE parameters").value();
				    // initialisation de maxDist    
				    maxDist =  _parser.getORcreateParam(unsigned(3), "maxDist", "Maximal number of predicate per station", 'D', "DAE parameters").value();
				    
				   // initialisation de la distance 
				     distance =  _parser.getORcreateParam(unsigned(100), "distance", "Number atoms per station", 'a', "DAE parameters").value();  
					//initialisation de avgNbAdditionalAtChange					
				    avgNbAdditionalAtChange = _parser.getORcreateParam(1.5, "avgNbAtChange", "Avg number of changes fir each predicates in atInit (>1)", '\0', "DAE parameters").value() - 1; 
				    if (avgNbAdditionalAtChange < 0)
				      throw std::runtime_error("avgNbAtChange must be > 1");
		             //initialisation de  pAddRemovePredicate		
				    pAddRemovePredicate =_parser.getORcreateParam(0.2, "pAddRemovePredicate", "Probability to change the 'present' state in init and mutation", '\0', "DAE parameters").value();
				     //initialisation de  pAddRemovePredicate		
				    pMeanAveragePerMutation = _parser.getORcreateParam(0.2, "pMeanAveragePerMutation", "Mean probability to change fluent of the  state in mutation", 'A', "DAE parameters").value();
					//initialisation de la penalite		
				    penalite = _parser.getORcreateParam(1e+60, "penalite", "Penality", 'p', "DAE parameters").value();
				    stationMax = _parser.getORcreateParam(20, "stationMax", "nombre de stations maximum autorisees", 'm', "DAE parameters").value();
				  	runTime =  _parser.getORcreateParam(1800, "runTime", "run duration", 't', "DAE parameters").value();
				   	 resultFile=  _parser.getORcreateParam(std::string("result"), "fileResult", "name of output file", 'r', "Output").value();
					//initialisation de minSize 			    			    
				    //minSize = minSizeParam.value();
				     //initialisation de nbPred
				    nbPred =_parser.getORcreateParam(unsigned(0), "nbPred", "nombre de predicat dans variables globales", 'P', "DAE parameters").value();
				    // initialisation du nombre maximum de backtracks autorises
				    maxb =_parser.getORcreateParam(std::string("2000"), "maxb", "nombre maximum de backtracks autorises", '\0', "DAE parameters").value();
				    // initialisation du fichier contenant la definition du domaine
				    NameObjects= _parser.getORcreateParam(std::string("object.pddl"), "f", "name of objects file", 'f', "Input").value();
				    // initialisation de fichier contenant la definition des objets
				    NameDomaine= _parser.getORcreateParam(std::string("domain.pddl"), "o", "name of domain file ", 'o', "Input").value();
				    //initialisation de atMutRatePerGoal
				    atMutRatePerGoal= _parser.getORcreateParam(0.1, "atMutRatePerGoal", "Mutation Rate per Patial Goal", '\0', "DAE parameters").value();
				    //lecture des predicats a utilise
				    std::string buf=  _parser.getORcreateParam(std::string("at/"), "liste", "liste des predicats", '\0', "DAE parameters").value(); 
				    // extraction des predicats utiles pour la construction des gares intermediaire
				   	precision=_parser.getORcreateParam(0.00001, "precision", "Validator precision", 'v', "DAE parameters").value();
				    cout << " \n fichier domaine " << NameDomaine << " \n fichier  des Objets " << NameObjects << "\n";
				    // construction de l'objet CPT -v 0 -M
				    std::string tmp(" ");
				    if(debug)tmp+=" -M -o ";
				    else tmp+="-v 0 -M -o ";
				    tmp +=NameDomaine + " -f " + NameObjects + " -maxb " + maxb + " -gs -n";
				    cptMakeProblem((char *)tmp.c_str());
					CPTinformationDAE();
					std::string res ("result");
					if(strcmp(resultFile.c_str(),res.c_str())==0)resultFile=NameObjects+".out";
					//maxb =world_size() / 10000;
					setMaxb(atoi(maxb.c_str()));
				    // mode debug
				    if (debug){
				    	CPTFluentArray init = cptGetInit();
						CPTFluentArray goal = cptGetGoal();
						
						cout << "\nINIT STATE\n\n";
						printFluentArray(init);
						
						cout << "\nGOAL STATE\n\n";
						printFluentArray(goal);
						cout << "initialisation des variables globales \n";			    	
				    	cout << "maxsize " << maxSize << " \n maxdist    " << maxDist << "\n";
				    	cout << "avg changes + " <<  avgNbAdditionalAtChange;
				    	cout << " \n add remove pre " <<  pAddRemovePredicate << "\n";
				    	cout << "min size " << minSize << " \n nb pred " << nbPred ;
				   		cout << " fichier domaine " << NameDomaine << " \n fichier  des Objets " << NameObjects << "\n";
				    
				    	printAttributeSpaces();
				    }
				    CPTFluentArray goal = cptGetGoal();
					stationMax=2*goal.nb;
				  	printDiscriminantAttribute();
				   	cout << "maxsize " << maxSize << " \n maxdist    " << distance << "\n";
				   
			  		 
			  		
						    
			}
			
			/**
			 * Definition des accesseurs  
			 */
			 
			 
			 /**
			  * @fn unsigned getDebug()
			  *@brief accesseur de la variable debug
			  */
			unsigned getDebug(void)
			{
				return debug;
			}
			double getPrecision(void)
			{
				return precision;
			}
			
			/**
			 * @fn unsigned int getMaxSize()
			  *@brief accesseur de la variable maxSize
			  */
			unsigned int getMaxSize(void)
			{
				return maxSize;
			}
			
			/**
			 * @fn unsigned int getMaxb()
			  *@brief accesseur de la variable maxSize
			  */
		 int getMaxb(void)
			{
				return maxbks;
			}
			
			void setMaxb(size_t newBks)
			{
				maxbks=newBks;
			}
			
			/**
			 * @fn unsigned int getMaxDist()
			  *@brief accesseur de la variable maxDist
			  */
			unsigned int getMaxDist(void){
				return  maxDist;
			}
			
		 	
		 	/**
		 	 * @fn unsigned int getMinSize()
			  *@brief accesseur de la variable minSize
			  */
			unsigned int getMinSize(void)
			{
				return minSize;
			}
			
			/**
			 * @fn unsigned getNbPred()
			  *@brief accesseur de la variable nbPred
			  */
			unsigned getNbPred(void)
			{
				return nbPred;
			}
			
			/**
			 * @fn double getAvgNbAdditionalAtChange()
			  *@brief accesseur de la variable avgNbAdditionalAtChange
			  */
			double getAvgNbAdditionalAtChange(void)
			{
				return avgNbAdditionalAtChange;
			}
			/**
			 * @fn double getPAddRemovePredicate()
			  *@brief accesseur de la variable pAddRemovePredicate
			  */
			double getPAddRemovePredicate(void)
			{
				return pAddRemovePredicate;
			}
			/**
			 * @fn double getPenalite()
			  *@brief accesseur de la variable penalite
			  */
			double getPenalite(void)
			{
				return penalite;
			}
			
			/**
			  *@brief accesseur a la taille de la liste des predicats
			  */
			unsigned getPredicateSize(){
				return NamePredicates.size();
			}
			/**
			  *@brief accesseur a la distance entre deux stations
			  */
			unsigned getDistance(){
				return distance;
			}
			
			/**
			  *@brief accesseur a un predicat de la liste des predicats
			  */
			
			std::string getPredicate(size_t pos){
				return NamePredicates[pos];
			}	
			
			void printOn(ostream& os) const
				{
				  
				  os << "--Domaine="<<NameDomaine<<"\n";
				  os << "--debug="<<debug<<"\n";
				  os << "--minSize="<<minSize<<"\n";
				  os << "--maxSize="<<maxSize<<"\n";
				  os << "--maxDist="<<maxDist<<"\n";
				  os << "--nbPred="<<nbPred<<"\n";
				  os << "--maxb="<<maxb<<"\n";
				  os << "--Objets="<<NameObjects<<"\n";
				  os << "--atMutRatePerGoal="<<atMutRatePerGoal<<"\n";
				  os << "--liste=";
				  for(size_t i=0;i<NamePredicates.size();i++)
				  {
				  	os << NamePredicates[i].c_str() << "/";
				  }
				  os<<"\n";
				  os << "--avgNbAtChange="<<avgNbAdditionalAtChange<<"\n";
				  os << "--pAddRemovePredicate="<<pAddRemovePredicate<<"\n";
				  os << "--penalite=" << penalite <<"\n";
				}
				
				
			void readFrom(istream& is)
	      {
		
			
	      }	
	      
	      void setInitialState()
	      {
	      	
	      }
	      
	      std::vector<unsigned>& getInitialState()
	      {
	      	return InitialState;
	      }
	      double getpMeanAveragePerMutation()
	      {
	      	return pMeanAveragePerMutation;
	      }
					
			
};

/**
 * \var global
 * @brief global : instance de la classe variable globale(VariablesGlobales) utilisee 
 * par toutes les methodes demandant un acces aux parametres
 * elle est initialisee qu'une seule fois dans toute l'application
 */
VariablesGlobales global;
#endif /*INDIVIDUDISCRIMINANT_H_*/
